#include <stdlib.h>
#include <ctype.h>

/*
 * Programa efectuado para comunicar essencialmente com MODEMS.
 *
 *  Este programa foi desenvolvido para uso escolar no ISEL (Centro de Calculo)
 * durante as aulas praticas de Redes de Comunicacao de Dados do Bacharelato
 * e Licenciatura.
 *
 *  Vitor Vaz da Silva (c) 31 Nov 1990
 *
 * Alteracoes 23 Mar 1991
 * - comentarios
 * - BaudTable e rotinas que a usam.
 *
 * Alteracoes 25 Mar 1991
 * - porta serie 1 e 2
 *
 * Alteracoes 09 Abr 1991
 * - DEL para permitir abortar programas em XENIX.
 *
 * Alteracoes 18 Abr 1991
 * - TX_READY na rotina SendCar
 *
 * Alteracoes 01 Mai 1991
 * - TestFisico para permitir fazer debuging `a interface fisica
 *
 * Alteracoes 11 Jul 1991
 * - Controlo de fluxo por software
 *
 * Versao 2.1 a partir de 12/10/1991
 *
 * Versao 3.0 a partir de 27/01/1992
 *
 * Alteracoes 27 Jan 1992
 * - Em conjunto com Walter Vieira:
 * - Modulo printer.c com a hipotese de emulacao de uma impressora
 * - Utilizacao de 2 portas serie em simultaneo
 * - Nova opcao no menu, para simular uma impressora
 *
 * Alteraes 29 Abril 2004
 * - Para usar uma vez mais este programa com fins educativos:
 *   - retirei o mdulo de simulao da impressora.
 *   - o nmero de portas em serial.h foi colocado como sendo 1
 *   - envio e recepo de ficheiros sem qualquer tipo de controlo de erro
 *   - visualizar ficheiros em hexa
 * - Esta verso foi compilada para a ANFEI/ATEC Vitor Vaz da Silva
 *
 * Alteraes 07 Junho 2005
 * - Acrescentado opo de repetio de uma tecla para permitir a
 *    visualizao num osciloscpio.
 *
 * ATENCAO:
 *    Este programa foi desenvolvido no TurboC, em modelo small, e devido `a
 * utilizacao da rotina de interrupt e' necessario, segundo o help do TurboC,
 * alterar as seguintes opcoes:
 *
 *  - Test stack overflow    OFF
 *  - Use register variables OFF
 *       Esta ultima no entanto nao faz muito sentido e podera' ficar a ON ...
 *
 *
*/

#include <stdio.h>
#include <conio.h>
#include <fcntl.h>
#include <signal.h>
#include <dos.h>
#include "tipos.h"
#include "serial.h"
#include "eco.h"
#include "ecolib.h"


char FileConf[80] ="ECO.CFG";    /* Ficheiro de configuracao */
FILE *Fconf;                     /* Handler do ficheiro de configuracao */
int LocalEcho=FALSE;             /* O eco de cada caracter enviado  feito localmente */
int Interpretar=TRUE;            /* os caracteres so interpretados ex: CR */

char DownLoading=FALSE;		 /* indica que se est a escrever num ficheiro */
FILE *DownFile;                  /* identificador do ficheiro a ser excrito */
char UpLoading=FALSE;            /* indica que se est a ler de um ficheiro */
FILE *UpFile;                    /* identificador do ficheiro a ser lido */

/*
 * Caracteres utilizados pela interface utilizador - eco
 *
*/
#define ALT_B     0x30
#define ALT_C     0x2e
#define ALT_D     0x20
#define ALT_E     0x12
#define ALT_F     0x21
#define ALT_G     0x22
#define ALT_H     0x23
#define ALT_I     0x17
#define ALT_K     0x25
#define ALT_L     0x26
#define ALT_N     0x31
#define ALT_O	  0x18
#define ALT_P     0x19
#define ALT_Q     0x10
#define ALT_R     0x13
#define ALT_S     0x1f
#define ALT_T     0x14
#define ALT_U     0x16
#define ALT_V     0x2F
#define ALT_1     0x78
#define ALT_2     0x79
#define DEL_K     0x53
#define FUNC_KEY  0x00
#define F1_KEY    0x3b
#define CTRL_C    0x03
#define DEL       0x7f
#define ESC	  0x1B

/*
 *
 * Teclas de cursor
 *
 *
*/
#define CURSOR_UP    'H'
#define CURSOR_DOWN  'P'
#define CURSOR_LEFT  'M'
#define CURSOR_RIGHT 'K'

#define CTRL_P 0x10            /* Control P => Print Screen ; DLE para PAD */

int Porta=0;                   /* Assumir o funcionamento com a porta 0 COM1 */
/*
 * Parity
 *
 * Mosta e permite alterar o tipo de paridade utilizada na comunicacao.
 *
 *
*/
int Parity(int mode)
{
UCHAR parity;
int par;

  parity=inportb(BASE(Porta) + LIN_CTRL);
  if(mode){
    if(parity & PENABLE)
    switch(parity & PMASK){
      case PEVEN: printf("\nEVEN\n"); break;
      case PODD:  printf("\nODD\n");  break;
    }
    else printf("\nNONE\n");
    printf("\n1- NONE\n2- EVEN\n3- ODD\n0- NAO ALTERAR\nOPCAO-> ");
    par=getch()-'0';
  }else par=fgetc(Fconf)-'0';
  if(par<=0 || par>3) return;
  if(par==1) outportb(BASE(Porta) + LIN_CTRL, parity & ~PENABLE);
  else outportb(BASE(Porta) + LIN_CTRL, parity & ~PMASK | PENABLE | (par==2 ? PEVEN : PODD) );
}

/*
 * NumBits
 *
 *  Mostra e permite alterar o numero de bits utilizados na comunicacao.
 *
 *
*/
NumBits(mode)
int mode;
{
UCHAR num;
int n;

  num=inportb(BASE(Porta) + LIN_CTRL);
  if(mode){
    switch(num & WLMASK){
      case WL5:  printf("\n5"); break;
      case WL6:  printf("\n6"); break;
      case WL7:  printf("\n7"); break;
      case WL8:  printf("\n8"); break;
    }
    printf(" BITS\nNumero de Bits (0 NAO ALTERAR) -> ");
    n=getch()-'0';
  }else n=fgetc(Fconf)-'0';
  if(n>=5 && n<=8) outportb(BASE(Porta) + LIN_CTRL,num & ~WLMASK | (n-5)& WLMASK);
}

/*
 * BaudRate
 *
 * Mostra e permite alterar o baud rate utilizado na comunicacao.
 *
 *
*/
BaudRate(mode)
int mode;
{
long baud;

  baud=GetBaud(Porta);
  if(mode){
    if(baud) printf("\nBit Rate %ld\n",baud);
    else printf("Unknown Bit Rate\n");
    printf("1- 75\n2- 150\n3- 300\n4- 600\n5- 1200\n6- 2400\n7- 4800\n");
    printf("8- 9600\n9- 19200\n:- 38400\n0- NAO ALTERAR\nOpcao -> ");
    baud=getch()-'0';
  }
  else baud=fgetc(Fconf)-'0';
  if(baud>=1 && baud<=10) SetBaud(BaudTable[baud - 1].val,Porta);
}

/*
 * Status
 *
 * Mostra o estado dos sinais do protocolo RS232-C
 *
 *
*/
Status()
{
int porta;

  printf("\nPorta selecionada %d\n",Porta+1);
  printf("DSR ");if(! DSR_ACTIVE(Porta)) printf("NAO ");printf("ACTIVO\n");
  printf("CTS ");if(! CTS_ACTIVE(Porta)) printf("NAO ");printf("ACTIVO\n");
  printf("DCD ");if(! DCD_ACTIVE(Porta)) printf("NAO ");printf("ACTIVO\n");
  printf("RI  ");if(!  RI_ACTIVE(Porta)) printf("NAO ");printf("ACTIVO\n");
  printf("RTS ");if(! RTS_ACTIVE(Porta)) printf("NAO ");printf("ACTIVO\n");
  printf("DTR ");if(! DTR_ACTIVE(Porta)) printf("NAO ");printf("ACTIVO\n");
  if(FullDuplex[Porta]) printf("FULL DUPLEX\n"); else printf("HALF DUPLEX\n");
  for(porta=0;porta<PORTAS;porta++){
    if(!UartPresente(porta)) continue;
    printf("Porta %d %d caracteres\n",porta+1,NumCaracteres(porta));
  }
}

/*
 *Address()
 *
 * Altera o endereco da porta serie e procede a sua inicializacao.
 *
 *
*/
Address(mode)
int mode;
{
char c;

  if(mode){
    printf("\nPorta 1 a %d",PORTAS);
    c=getch();
  }else c=fgetc(Fconf);
  c-='1';
  if(c<PORTAS && UartPresente(c)) Porta=c;
  else printf("O ACE nao esta' presente ou esta' estragado!\n");
}

/*
 * SendBreak
 *
 * Permite o envio do caracter BREAK.
 *
 *
*/
SendBreak()
{
UCHAR save;
int i;

  save=inportb(BASE(Porta) + LIN_CTRL);
  outportb(BASE(Porta) + LIN_CTRL, save | BRK);
  printf("\n*** SENDING BREAK -- ");
  getch();
  outportb(BASE(Porta) + LIN_CTRL, save);
  printf(" STOPED ***\n");
}

/*
 *DownLoad()
 *
 * Abre um ficheiro para escrita.
 * Os caracteres recebidos pela porta srie enquanto se estiver no modo
 * terminal so escritos neste ficheiro.
 * Evocar pela segunda vez este comando, termina-se a escrita e fecha-se
 * o ficheiro.
 *
*/
DownLoad()
{
char nome[30];

  if(DownLoading){
    fclose(DownFile);
    DownLoading=FALSE;
  }
  else{
    printf("<ECO Download file -> ");
    scanf("%s",nome);getchar();
    if((DownFile=fopen(nome,"r"))!=NULL){
      printf("O ficheiro existe, quer continuar? ");
      fclose(DownFile);
      if(tolower(getchar())!='s') return;
    }
    if((DownFile=fopen(nome,"w+"))==NULL) printf("Erro ao criar o ficheiro\n");
    else{
      DownLoading=TRUE;
    }
  }
}

/*
 *UpLoad()
 *
 * Abre um ficheiro para leitura, caso exista.
 * Enquanto se estiver no modo terminal o ficheiro  enviado.
 * Evocar pela segunda vez este comando, termina a leitura e fecha-se
 * o ficheiro, caso este ainda no tenha chegado ao fim.
 *
*/
UpLoad()
{
char nome[30];

  if(UpLoading){
    fclose(UpFile);
    UpLoading=FALSE;
  }
  else{
    printf("<ECO Upload file -> ");
    scanf("%s",nome);getchar();
    if((UpFile=fopen(nome,"r"))==NULL){
      printf("O ficheiro no existe.");
    }
    else UpLoading=TRUE;
  }
}

/*
 * printavel
 *
 * Devolve o caracter car caso seja printvel ou caso contrrio o '.'
 *
 *
*/
printavel(car)
char car;
{
  return(isprint(car) ? car : '.');
}


/*
 * MostraLinha
 *
 * Imprime no ecran o nmero linhas e o contedo do buffer buf em hexa e
 * em ASCII caso seja possvel. A dimenso de buf  n
 *
 *
*/
MostraLinha(linhas, buf, n)
int n,linhas;
unsigned char buf[16];
{
int i;


    printf("%04X ", linhas);
    for(i=0;i!=n;) printf("%02X ", buf[i++]);
    for(;i++<16;) printf("   ");
    printf("   ");
    for(i=0;i!=n;) printf("%c", printavel(buf[i++]));
    printf("\n");
}

/*
 * ViewFile
 *
 * Mostra o contedo de um ficheiro em binrio e ASCII.
 * Permite controlar o fluxo de visualizao atravs dos seguintes comandos:
 * b - blocos de 8 linhas
 * l - linha a linha
 * c - de um modo contnuo
 *     qualquer caracter pra a visualizao quando em modo contnuo
 * ESC - interrompe a vizualizao do ficheiro
 *
 *
*/
ViewFile()
{
int n,i, linhas=0, cont=0, dim=0x07;
unsigned char buf[16];
char nome[30];
FILE *fich;

  printf("<ECO Ver ficheiro -> ");
  scanf("%s",nome);getchar();
  if((fich=fopen(nome,"rb"))==NULL){
    printf("O ficheiro no existe.");
  }
  else{
    while((n=fread(buf, 1, sizeof(buf), fich))!=0){
      MostraLinha(linhas, buf, n);
      if(!(++linhas & dim)){
	if(!cont) {
	  while(!kbhit());
	  switch(getch()){
	    case 'c': cont=1; break;
	    case ESC:
		fclose(fich);
		return;
	    case 'l': dim= 0x00; break;
	    case 'b': dim= 0x07; break;
	  }
	}
	else cont = !kbhit();
      }
    }
    fclose(fich);
  }
}

/*
 * RepeteCar
 *
 * L um caracter e escreve-o repetidamente para a porta de sada.
 * Termina assim que se prime a tecla 'q'. Qualquer outra tecla  enviada.
 *
*/
RepeteCar()
{
int car;
    printf("caracter a ecoar: ");
    car=getch();
    printf(" prima a tecla 'q' para terminar, qualquer outra para ecoar: ");
    do{
      if(kbhit()){
	if((car=getch())=='q') break; else putchar(car);
      }
      if (SendCar(car,Porta)!=0) putchar('.');
    }while(1);
}

/*
 * Terminal
 *
 * Coloca o programa a servir de tty. Permite a execucao dos comandos mas
 * precedendo-os de <ALT>.
 *
*/
Terminal(mode)
int mode;
{
int car;
int val;
UCHAR ExitTerminal,Sair;

  putchar('\n');
  ExitTerminal=1;
  Sair=1;
  do{
    if((val=GetCar(Porta))!=-1){
      if(Interpretar){
	if(val==0x0a){
	  putch(0x0a);
	}else putchar(val);
      } else putch(val);
      if(DownLoading){
	fputc(val,DownFile);
      }
    }
    if(UpLoading){
      if((car=fgetc(UpFile)) == EOF){
	UpLoading=0;
	fclose(UpFile);
      }
      if(SendCar(car,Porta)==0 && LocalEcho) putchar(car);
    }
    TratarSpecial(Porta);
    if(kbhit()){
      if((car=getch())!=0){
	if(SendCar(car,Porta)==0 && LocalEcho) putchar(car);
      }
      else{
	switch(car=getch()){
	  case ALT_S:  Status();              break;
	  case ALT_B:  BaudRate(mode);        break;
	  case ALT_P:  Parity(mode);          break;
	  case ALT_N:  NumBits(mode);         break;
	  case ALT_D:  DtrChange(mode);       break;
	  case ALT_F:  HalfFullDuplex(mode);  break;
	  case ALT_G:  DownLoad();            break;
	  case ALT_U:  UpLoad();              break;
	  case ALT_V:  ViewFile();            break;
	  case ALT_L:  LocalEchoChg(mode);    break;
	  case ALT_R:  RtsChange(mode);       break;
	  case ALT_1:  Sistema(mode);         break;
	  case ALT_2:  SendBreak();           break;
	  case ALT_E:  SendCar(CTRL_P,Porta); break;
	  case F1_KEY: Help();                break;
	  case ALT_C:  Config();              break;
	  case ALT_K:  SendCar(CTRL_C,Porta); break;
	  case ALT_O:  RepeteCar();	      break;
	  case ALT_Q:
		Sair=0;
	  case ALT_T:  ExitTerminal=0;
               break;
	}
	/* Para utilizar o vi ; feito (mais ainda) `a pressao ... */
	switch(car){
	  case DEL_K:        SendCar(DEL,Porta);    break;
	  case CTRL_C:       SendCar(CTRL_C,Porta); break;
	  case CURSOR_UP:    SendCar(0x1B,Porta); SendCar('k',Porta); break;
	  case CURSOR_DOWN:  SendCar(0x1B,Porta); SendCar('j',Porta); break;
	  case CURSOR_LEFT:  SendCar(0x1B,Porta); SendCar('h',Porta); break;
	  case CURSOR_RIGHT: SendCar(0x1B,Porta); SendCar('l',Porta); break;
	  default: putchar('#');
	}
      }
    }
  }while(ExitTerminal);
  return(Sair);
}

/*
 * Config
 *
 * Mostra os parametros utilizados na comunicacao.
 *
 *
*/
Config()
{
UCHAR port;
int baud;

  printf("\nPARITY - ");
  port=inportb(BASE(Porta) + LIN_CTRL);
  if(port & PENABLE)
  switch(port & PMASK){
    case PEVEN: printf("EVEN\n");
      break;
    case PODD:  printf("ODD\n");
      break;
  }
  else printf("NONE\n");
  printf("BITS   - ");
  switch(port & WLMASK){
    case WL5:  printf("5");
      break;
    case WL6:  printf("6");
      break;
    case WL7:  printf("7");
      break;
    case WL8:  printf("8");
      break;
  }
  baud=GetBaud(Porta);
  if(baud) printf("\nBit Rate %d\n",baud);
  else printf("\nUnknown Bit Rate\n");
}

/*
 * DtrChange
 *
 * Permite alterar o estado do sinal DTR.
 *
 *
*/
DtrChange(mode)
int mode;
{
char c;
  if(mode){
    printf("\nDTR %s ACTIVO ; 1- ACTIVAR   0- DESACTIVAR ",DTR_ACTIVE(Porta) ? "" : "NAO");
    c=getch();
  }
  else c=fgetc(Fconf);
  if(c=='1') DTR_ON(Porta);
  else if(c=='0') DTR_OFF(Porta);
}

/*
 * RtsChange
 *
 * Permite alterar o estado do sinal RTS.
 *
 *
*/
RtsChange(mode)
int mode;
{
char c;
  if(mode){
    printf("\nRTS %s ACTIVO ; 1- ACTIVAR   0- DESACTIVAR ",RTS_ACTIVE(Porta) ? "" : "NAO");
    c=getch();
  }else c=fgetc(Fconf);
  if(c=='1') RTS_ON(Porta);
  else if(c=='0') RTS_OFF(Porta);
}

/*
 * HalfFullDuplex
 *
 * Permite alterar o tipo de comunicacao half/full duplex.
 *
 *
*/
HalfFullDuplex(mode)
int mode;
{
char c;
  if(mode){
    printf("\n%s DUPLEX    1-FULL    0-HALF ",FullDuplex ? "FULL" : "HALF");
    c=getch();
  } else c=fgetc(Fconf);
  if(c=='1') FullDuplex[Porta]=TRUE;
  else if(c=='0') FullDuplex[Porta]=FALSE;
}

/*
 * LocalEcoChg
 *
 * Permite alterar o tipo de eco, local ou remoto.
 *
 *
*/
LocalEchoChg(mode)
int mode;
{
char c;
  if(mode){
    printf("\nEco %s    1-LOCAL    0-REMOTO ",LocalEcho ? "Local" : "Remoto");
    c=getch();
  } else c=fgetc(Fconf);
  if(c=='1') LocalEcho=TRUE;
  else if(c=='0') LocalEcho=FALSE;
}

/*
 * InterpretChar
 *
 * Permite alterar o tipo de apresentacao dada aos caracteres nao printaveis.
 *
 *
*/
InterpretChar(mode)
int mode;
{
char c;
  if(mode){
    printf("\nInterpreta caracteres de control %s     1-SIM   0-NAO ",
     Interpretar ? "Sim" : "Nao");
    c=getch();
  } else c=fgetc(Fconf);
  if(c=='1') Interpretar=TRUE;
  else if(c=='0') Interpretar=FALSE;
}

/*
 * ControloFluxo
 *
 * Utilizar controlo de fluxo utilizando XON/XOFF
 *
 *
*/
ControloFluxo(mode)
int mode;
{
char c;
  if(mode){
    printf("\nControlo de Fluxo %s     1-SIM   0-NAO ",
     CtrlFluxo ? "Sim" : "Nao");
    c=getch();
  } else c=fgetc(Fconf);
  if(c=='1') CtrlFluxo[Porta]=1;
  else if(c=='0') CtrlFluxo[Porta]=0;
}


/*
 * Sistema
 *
 * Permite evocar apenas um comando do MS-DOS.
 *
 *
*/
Sistema(mode)
int mode;
{
char arr[80];
char *s = arr;

  if(mode){
    printf("DOS> ");
    while((*s++=getchar())!='\n');
  }
  else  while((*s++=fgetc(Fconf))!='\n');
  *(--s)=0;
  system(arr);
}


/*
 * TestFisico
 *
 *  Com o auxilio de curto circuitos na ficha D, permite testar o funcionamento
 * da interface fisica;
 *
*/
TestFisico()
{
#define DSR_ERR 0x01
#define RI_ERR  0x02
#define CTS_ERR 0x04
#define DCD_ERR 0x08
#define RX_ERR  0x10
UCHAR c=0;

  if(NaoExisteACE(Porta)){
    printf("\nO ACE nao esta' presente ou esta' estragado!\n");
    return;
  }
  outportb(BASE(Porta) + LIN_CTRL, inportb(BASE(Porta) + LIN_CTRL) | WL8 );
  DTR_OFF(Porta); delay(10);
  if(DSR_ACTIVE(Porta)) c |= DSR_ERR;
  if(RI_ACTIVE(Porta)) c |= RI_ERR;
  DTR_ON(Porta); delay(10);
  if(! DSR_ACTIVE(Porta)) c |= DSR_ERR;
  if(! RI_ACTIVE(Porta)) c |= RI_ERR;
  RTS_OFF(Porta); delay(10);
  if(CTS_ACTIVE(Porta)) c |= CTS_ERR;
  if(DCD_ACTIVE(Porta)) c |= DCD_ERR;
  RTS_ON(Porta); delay(10);
  if(! CTS_ACTIVE(Porta)) c |= CTS_ERR;
  if(! DCD_ACTIVE(Porta)) c |= DCD_ERR;
  while(GetCar(Porta) != -1) ;
  SendCar(0x55,Porta);
  sleep(1);
  if((GetCar(Porta)) != 0x55) c |= RX_ERR;
  SendCar(0xAA,Porta);
  sleep(1);
  if((GetCar(Porta)) != 0xAA) c |= RX_ERR;
  if(!c) printf("\nInterface fisica em condicoes");
  if( !(c ^ (DSR_ERR | DCD_ERR | RI_ERR | RX_ERR | CTS_ERR)))
    printf("\nVerifique se a ficha com os curto-circuitos esta bem posta!");
  if( c & DSR_ERR) printf("\nErro nos sinais DTR - DSR");
  if( c & DCD_ERR) printf("\nErro nos sinais CTS - DCD");
  if( c & RI_ERR ) printf("\nErro nos sinais DTR - RI");
  if( c & RX_ERR ) printf("\nErro nos sinais TX  - RX");
  if( c & CTS_ERR) printf("\nErro nos sinais CTS - RTS");
}

/*
 * Menu
 *
 * Menu
 *
 *
*/
Menu()
{
  clrscr();
  LogOn();
  printf("\nF1 help         ? menu         ! system\nc config\nn numero de bits\np parity\n");
  printf("b baud\nd DTR change\nr RTS change\nf half/full duplex\ns status\n");
  printf("g guarda em ficheiro\nl local echo\ny controlo fluxo\ni interpretar\n");
  printf("u envia ficheiro\nv ver ficheiro\nz codigo da tecla\n");
  printf("o repete caracter\nt terminal\nx testfisico\nq quit\n");
}

/*
 * Help
 *
 * Mostra a funcao dos comandos acessiveis.
 *
 *
*/
Help()
{
  clrscr();
  LogOn();
  printf("*********** HELP ***********\n");
  printf("Comando q QUIT             ");
  printf("Abandonar este programa\n");
  printf("Comando s STATUS           ");
  printf("Mostra o estado dos sinais RS232-C ; V.24\n");
  printf("Comando ! SYSTEM           ");
  printf("Permite executar um comando do sistema operativo\n");
  printf("Comando c CONFIG           ");
  printf("Mostra a configuracao actual\n");
  printf("Comando l LOCAL ECHO       ");
  printf("Eco feito localmente, mas apos o envio para o modem!\n");
  printf("Comando d DTR CHANGE       ");
  printf("Alterar o estado do sinal DTR\n");
  printf("Comando r RTS CHANGE       ");
  printf("Alterar o estado do sinal RTS\n");
  printf("Comando b BIT RATE         ");
  printf("Mostra o bit rate actual e permite modifica-lo.\n");
  printf("Comando p PARIDADE         ");
  printf("Tipo de paridade a usar\n");
  printf("Comando n NUMERO DE BITS   ");
  printf("Numero de bits a transmitir nao contando a paridade\n");
  printf("Comando f HALF\\FULL DUPLEX ");
  printf("Determina o modo de funcionamento dos sinais RTS/CTS\n");
  printf("Comando t TERMINAL         ");
  printf("Permite colocar este programa como emulador de\n");
  printf("terminal sendo no entanto permitido o uso dos comandos precedidos de ALT:\n");
  printf("ex.  ALT 1 Equivalente ao comando ! System\n");
  printf("     ALT K - Para o envio do caracter 0x03 (CTRL C)\n");
  printf("     ALT E - Para o envio do caracter 0x10 (CTRL P)\n");
  printf("     F1    - Este Help\n");
  printf("     @     - Envia Break\n");
  printf("     ALT T - Retorna ao modo de comando\n");
  printf("Comando g seguido do nome do ficheiro onde fica guardado o que se recebe em modo terminal.\n");
  printf("Comando u seguido do nome do ficheiro a enviar em modo terminal.\n");
  printf("Comando v seguido do nome do ficheiro a visualizar em hexa e ascii\n");
  printf("Comando z identificao do cdigo (2) da tecla primida em hexa.\n");
  printf("***************************");
}

/*
 * Executar
 *
 * Este procedimento executa os comandos dados pelo ficheiro de configuracao ou
 * pelo utilizador.
 *
 *
*/
Executar(c,mode)
char c;
int mode;
{
int Sai;

  switch(c){
    case 'a': Address(mode);           break;
    case 't': Sai=Terminal(mode);      break;
    case 'c': Config();                break;
    case 'd': DtrChange(mode);         break;
    case 'r': RtsChange(mode);         break;
    case 'f': HalfFullDuplex(mode);    break;
    case 's': Status();                break;
    case 'b': BaudRate(mode);          break;
    case 'p': Parity(mode);            break;
    case 'n': NumBits(mode);           break;
    case 'l': LocalEchoChg(mode);      break;
    case 'i': InterpretChar(mode);     break;
    case 'y': ControloFluxo(mode);     break;
    case '?': Menu();                  break;
    case '!': Sistema(mode);           break;
    case 'x': TestFisico();            break;
    case 'o': RepeteCar();	       break;
    case FUNC_KEY:
      if(getch()!=F1_KEY) break;
    case 'h':
      Help(mode);
      break;
    case 'z': /* Identificacao do codigo da tecla premida. Para debug */
      printf("<0x%x>",getch());
      printf("<0x%x>",getch());
      break;
    case 'e':
    case 'q': Sai=0;
      break;
    default:
      return(Sai);
  }
  putchar('\n');
  return(Sai);
}

/*
 * LogOn
 *
 * Mensagem inicial.
 *
 *
*/
LogOn()
{
  printf("\rRedes de Comunicacao de Dados\n");
  printf("Emulador de Terminal ECO - v4.0 VVS (c) CCISEL 1990 1991 @ 1997 ATEC 2004-2005\n");
  printf("Vitor Vaz da Silva http:\\tektonia.no.sapo.pt\n");
}

/*
 * main
 *
 * Main.
 *
 *  Caso exista o ficheiro de configuracao, sao executados os comandos nele
 * existentes, e de seguida os introduzidos pelo utilizador em modo de comando.
 *
*/
main(argc,argv)
int argc;
char *argv[];
{
char c;

  StartEco();
  if(Init()) printf("O ACE nao esta' presente ou esta' estragado!\n");
  clrscr();
  LogOn();
  if(argc>2){
    printf("usage: eco [filename]\n");
    exit(1);
  }
  if(argc==2){
    strcpy(FileConf,argv[1]);
  }
  if((Fconf=fopen(FileConf,"r"))!=NULL){
    while((c=fgetc(Fconf))!=-1){
      if(c==' ') while((c=fgetc(Fconf))==' ') ;
      if(c=='/' && (c=fgetc(Fconf))=='*'){
    c=' ';
    do{
      while((c!='*')&&(c>=0)){
        c=fgetc(Fconf);
      }
      if(c==-1) break;
      c=fgetc(Fconf);
    }while(c!='/');
    if(c!=-1) continue;
      }
      if(c==-1) break;
      Executar(c,0);
    }
  }
  Menu();
  do{
    printf("\r      \rECO> ");
    while(!kbhit()) TratarSpecial(Porta);
    c=tolower(getch());
    putch(c);
  }while(Executar(c,1));
  EndSession();
  EndEco();
}

