/*
 * Programa efectuado para comunicar essencialmente com MODEMS.
 *
 *  Este programa foi desenvolvido para uso escolar no ISEL (Centro de Calculo)
 * durante as aulas praticas de Redes de Comunicacao de Dados do Bacharelato
 * e Licenciatura.
 *
 *  Vitor Vaz da Silva (c) 31 Nov 1990
 *
 * Alteracoes 23 Mar 1991
 * - comentarios
 * - BaudTable e rotinas que a usam.
 *
 * Alteracoes 25 Mar 1991
 * - porta serie 1 e 2
 *
 * Alteracoes 09 Abr 1991
 * - DEL para permitir abortar programas em XENIX.
 *
 * Alteracoes 18 Abr 1991
 * - TX_READY na rotina SendCar
 *
 * Alteracoes 01 Mai 1991
 * - TestFisico para permitir fazer debuging `a interface fisica
 *
 * Alteracoes 11 Jul 1991
 * - Controlo de fluxo por software
 *
 * Versao 2.1 a partir de 12/10/1991
 *
 * Alteracoes 22 Nov 1991
 * - Timeout para utilizacao de algoritmos
 *
 * Versao 3.0 a partir de 27/01/1992
 *
 * Alteracoes 27 Jan 1992
 * - Em conjunto com Walter Vieira:
 * - Modulo printer.c com a hipotese de emulacao de uma impressora
 * - Utilizacao de 2 portas serie em simultaneo
 * - Nova opcao no menu, para simular uma impressora
 *
 * ATENCAO:
 *    Este programa foi desenvolvido no TurboC, em modelo small, e devido `a
 * utilizacao da rotina de interrupt e' necessario, segundo o help do TurboC,
 * alterar as seguintes opcoes:
 *
 *  - Test stack overflow    OFF
 *  - Use register variables OFF
 *       Esta ultima no entanto nao faz muito sentido e podera' ficar a ON ...
 *
*/

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <dos.h>
#include "tipos.h"
#include "serial.h"
#include "eco.h"


#define MOD_DIM   (2048)        /* Dimensao maxima do buffer */
#define HIGH_MARK 64            /* Valor de histeres para o buffer de recepcao */
#define LOW_MARK 128            /* Valor de histeres para o buffer de recepcao */

#define TIMEOUT  102400L        /* Valor de timeout na transmissao HDX */

static UINT put[PORTAS];        /* Indices de por nos buffer de recepcao */
static UINT get[PORTAS];        /* Indices de tirar nos buffer de recepcao */
static UCHAR Buffer[PORTAS][MOD_DIM];      /* buffers de recepcao */
void interrupt (*SaveIntr[PORTAS]) ();   /* variavel utilizada para guardar */
					 /* rotinas de interrupt anteriors */
typedef struct{
	void interrupt (*intr)();
	UCHAR mdm_intr;
} Context;
Context Save[PORTAS];
UCHAR SavedMask;			/* Estado dos interrupts no inicio */

static long Ticks;            /* Numero actual de ticks; 18.2 ticks/seg */
static BYTE TimerOn;          /* Estado do contador de temporizacao */

void interrupt (*SaveIntr1C) ();  /* Guardar a rotina de Ticks do DOS */

#if PORTAS == 2
ACE Uart[PORTAS] = { {PORT1,IRQPORT1,IRQMASK1},{PORT2,IRQPORT2,IRQMASK2} };
#else
ACE Uart[PORTAS] = { {PORT1,IRQPORT1,IRQMASK1} };
#endif

/*
 * Variaveis utilizadas para controlo de fluxo; histerese
 *
*/
#if PORTAS == 2
  char CtrlFluxo[PORTAS]={TRUE,TRUE};
  UCHAR HighWater[PORTAS]={FALSE,FALSE};
  UCHAR LowWater[PORTAS]={TRUE,TRUE};
  static UCHAR Special[PORTAS]={0,0};
#else
  char CtrlFluxo[PORTAS]={TRUE};
  UCHAR HighWater[PORTAS]={FALSE};
  UCHAR LowWater[PORTAS]={TRUE};
  static UCHAR Special[PORTAS]={0};
#endif
#define Distancia(a,b) ( ((b)-(a)) % MOD_DIM)


/*
 * Caracteres utilizados para controlo de fluxo.
 *
*/
#define XON    0x11              /* Control Q */
#define XOFF   0x13              /* Control S */

/*
 * Tabela de baud rates
 *
*/
#define BAUD_DIM   10
BAUD_TAB BaudTable[BAUD_DIM] = {
             {    75L, B75    },
             {   150L, B150   },
             {   300L, B300   },
             {   600L, B600   },
             {  1200L, B1200  },
             {  2400L, B2400  },
             {  4800L, B4800  },
             {  9600L, B9600  },
             { 19200L, B19200 },
             { 38400L, B38400 }
     };


/*
 * Variaveis de estado.
 *
*/
int FullDuplex[PORTAS];

/*****************************************************************************
                  Procedimentos
*****************************************************************************/

/*
 * UartPresente
 *
 *  Indica a presenca do UART associado `a porta indicada.
 *
 *
*/
int UartPresente(int porta)
{
  return(SaveIntr[porta] != NULL);
}

/*
 * SetBaud
 *
 * Altera o baud rate do 8250. O valor passado como parametro e' aquele pelo
 * qual se quer dividir o clock de 1.8432 MHz.
 *
*/
SetBaud(b,porta)
int b,porta;
{
UCHAR save;

  save=inportb(BASE(porta) + LIN_CTRL);
  outportb(BASE(porta) + LIN_CTRL, save | DLAB);
  outportb(BASE(porta) + LOW_BAUD, b );
  outportb(BASE(porta) + HIGH_BAUD, b >> 8);
  outportb(BASE(porta) + LIN_CTRL, save);
}

/*
 * GetBaud
 *
 * Devolve o valor do baud rate
 *
 *
*/
long GetBaud(porta)
int porta;
{
UCHAR save;
long baud;
int i;

  save=inportb(BASE(porta) + LIN_CTRL);
  outportb(BASE(porta) + LIN_CTRL, save | DLAB);
  baud= inportb(BASE(porta) + LOW_BAUD) + (inportb(BASE(porta) + HIGH_BAUD)<< 8) ;
  outportb(BASE(porta) + LIN_CTRL, save);
  for(i=0;i<BAUD_DIM;i++)
    if(BaudTable[i].val == baud) break;
  return(i >= BAUD_DIM ? 0L : BaudTable[i].baud);
}

/*
 * SendCar
 *
 *  Envia um caracter para o UART. No caso da comunicacao ser em half-duplex
 * O RTS e' pulsado, e espera-se por CTS...
 *
*/
SendCar(car,porta)
UCHAR car;
int porta;
{
double i;

  if(FullDuplex[porta]){
    while (!TX_READY(porta)) ;
    OutCar(car,porta);
    i=0;
  }
  else{
    RTS_ON(porta);
    for(i=0;i<TIMEOUT;i++){
      if(CTS_ACTIVE(porta)){
    while (!TX_READY(porta)) ;
    OutCar(car,porta);
    i=0;
    break;
      }
    }
    RTS_OFF(porta);
  }
  return(i);
}

/*
 * BufNaoVazio
 *
 * Indica o estado do buffer de recepcao.
 *
 *
*/
BufNaoVazio(porta)
int porta;
{
register int r;
  disable();
  r= put[porta];
  enable();
  return(r!=get[porta]);
}

/*
 * NumCaracteres
 *
 *  Devolve o numero de caracteres que se encontram no buffer de recepcao.
 *
 *
*/

int NumCaracteres(porta)
int porta;
{
  return(Distancia(get[porta],put[porta]));
}

/*
 * GetCar
 *
 *  Devolve um caracter lido do buffer de recepcao.
 *
 *
*/
GetCar(porta)
int porta;
{
UCHAR c;

  if(BufNaoVazio(porta)){
    c=Buffer[porta][get[porta]];
    get[porta]=(get[porta] + 1 ) % MOD_DIM;
    if(CtrlFluxo[porta] && !LowWater[porta] &&
			    Distancia(get[porta],put[porta])==LOW_MARK){
      Special[porta]=XON;
      LowWater[porta]=TRUE;
    }
    return(c);
  }else return(-1);
}

/*
 * TratarSpecial
 *
 * Proceder ao envio do caracter XON / XOFF
 *
 *
*/
TratarSpecial(porta)
int porta;
{
  if(Special[porta]){
    SendCar(Special[porta],porta);
    if(Special[porta] == XOFF) LowWater[porta]=FALSE; else HighWater[porta]=FALSE;
    Special[porta]=0;
  }
}


/*
 * RxIntr
 *
 *  Rotinas de interrupt de recepcao.
 *  Assim que for obtida a condicao de "buffer cheio", nao se colocam mais
 * caracteres no buffer, nao sendo assim possivel a sua recuperacao.
 *
*/

void RxIntr(porta)
int porta;
{
UINT buu;

  if(get[porta] != (buu = (put[porta] + 1) % MOD_DIM)) {
    Buffer[porta][put[porta]]=InCar(porta);
    put[porta] = buu;
  }else InCar(porta); /* perdem-se caracteres */
  if(CtrlFluxo[porta]){
    if(!HighWater[porta] && Distancia(put[porta],get[porta]) == HIGH_MARK ){
      HighWater[porta]=TRUE;
      Special[porta]=XOFF;
    }
  }
  outportb(PICADDR,EOI);
}

typedef void interrupt (* INTRFUNC)();
extern INTRFUNC IntrHandler[PORTAS];

/*
 * NaoExisteACE
 *
 *    Este procedimento detecta se existe uma port de input/output no
 * endereco BASE + LIN_CTRL
 *
*/
NaoExisteACE(porta)
int porta;
{
UCHAR save;

  save=inportb(BASE(porta) + LIN_CTRL);
  outportb(BASE(porta) + LIN_CTRL,0x55);
  if(inportb(BASE(porta) + LIN_CTRL)==0x55){
    outportb(BASE(porta) + LIN_CTRL,0xAA);
    if(inportb(BASE(porta) + LIN_CTRL)==0xAA){
      outportb(BASE(porta) + LIN_CTRL,save);
      return(0);
    }
  }
  outportb(BASE(porta) + LIN_CTRL,save);
  return(-1);
}

/*
 * Init
 *
 * Rotina de iniciacao.
 *
 *
*/
InitPortas()
{
int result=0;
int porta;

  for(porta=0;porta<PORTAS;porta++){
    if(NaoExisteACE(porta)){
      result|= 1 << porta;
      SaveIntr[porta]=NULL;
      continue;
    }
    get[porta]=0;
    put[porta]=0;
    Save[porta].mdm_intr=inportb(BASE(porta) + INTR_EN);
    SaveIntr[porta] = getvect(IRQPORT(porta));
    setvect(IRQPORT(porta),IntrHandler[porta]);
    outportb(BASE(porta) + INTR_EN,RS_ALL_INTR);
    SetBaud(B9600);
    outportb(BASE(porta) + LIN_CTRL,INIT_LIN_CTRL);
    outportb(BASE(porta) + MDM_CTRL,OUT2);
    outportb(BASE(porta) + INTR_EN,ST_RDA_INTR);
    outportb(MASKADDR,inportb(MASKADDR) & ~IRQMASK(porta));
    DTR_ON(porta);
    FullDuplex[porta]=TRUE;
  }
  return(result);
}

Init()
{
  signal(SIGINT,SIG_IGN);
  return(InitPortas());
}

/*
 * EndSession
 *
 * Rotina final.
 *
 *
*/
EndSession()
{
int porta;

  for(porta=0;porta<PORTAS;porta++){
    if(SaveIntr[porta] == NULL) continue;
    DTR_OFF(porta);
    outportb(MASKADDR,inportb(MASKADDR) | IRQMASK(porta));
    setvect(IRQPORT(porta),SaveIntr[porta]);
    outportb(BASE(porta) + INTR_EN,Save[porta].mdm_intr);
  }
  return(0);
}

/*
 * EndEco
 *
 * Rotina final. Repoe a mascara dos interrupts
 *
 *
*/
EndEco()
{
    outportb(MASKADDR,SavedMask);
}

/*
 * StartEco
 *
 * Rotina inicial. Guarda o estado dos interrupts.
 *
 *
*/
StartEco()
{
    SavedMask=inportb(MASKADDR);
}

/*
 * Timer
 *
 *   Rotina de interrupt utilizada para decrementar a variavel Ticks
 *
 *
*/
void interrupt Timer()
{
  if(Ticks) Ticks--;
  SaveIntr1C();          /* Pode la' estar algo pendurado! */
}

/*
 * InitTimer()
 *
 *  Procede as inicializacoes necessarias para a utilizacao do timer.
 *
 *
*/
InitTimer()
{
  Ticks=0L;
  TimerOn=FALSE;
  SaveIntr1C=getvect(0x1C);
  setvect(0x1C,Timer);
}

/*
 * EndTimer
 *
 *  Deixa o interrupt do timer como estava antes de se ter evocado o InitTimer.
 *
 *
*/
EndTimer()
{
  setvect(0x1C,SaveIntr1C);
}

/*
 * Start Timer
 *
 *  Comeca a contar o tempo.
 *
 *
*/
StartTimer(int timeout)
{
  disable();
  Ticks=timeout;
  enable();
  TimerOn=TRUE;
}

/*
 * StopTimer
 *
 *  Para o contador, e devolve o valor que faltava decrementar
 *
 *
*/
long StopTimer()
{
long l;
  disable();
  l=Ticks;
  Ticks=0L;
  enable();
  TimerOn=FALSE;
  return(l);
}

/*
 * TimerValue
 *
 *  Devolve o valor do contador "on fly".
 *
 *
*/
long TimerValue()
{
long l;
  disable();
  l=Ticks;
  enable();
  return(l);
}

/*
 * Timeout
 *
 *  Indica se o contador expirou ou nao.
 *
 *
*/
Timeout()
{
  if(TimerOn)
   if(TimerValue() == 0L){
     TimerOn=FALSE;
     return(TRUE);
   }
  return(FALSE);
}
